@@ -4,6 +4,7 @@
  * 
  * WvStream-based TCP connection class.
  */
+#include "wvtcplistener.h"
 #include "wvtcp.h"
 #include "wvistreamlist.h"
 #include "wvmoniker.h"
@@ -53,9 +54,10 @@
 #endif
 
 WV_LINK(WvTCPConn);
+WV_LINK(WvTCPListener);
 
 
-static IWvStream *creator(WvStringParm s)
+static IWvStream *creator(WvStringParm s, IObject*)
 {
     return new WvTCPConn(s);
 }
@@ -63,6 +65,20 @@ static IWvStream *creator(WvStringParm s)
 static WvMoniker<IWvStream> reg("tcp", creator);
 
 
+static IWvListener *listener(WvStringParm s, IObject *)
+{
+    WvConstStringBuffer b(s);
+    WvString hostport = wvtcl_getword(b);
+    WvString wrapper = b.getstr();
+    IWvListener *l = new WvTCPListener(hostport);
+    if (l && !!wrapper)
+	l->addwrap(wv::bind(&IWvStream::create, wrapper, _1));
+    return l;
+}
+
+static WvMoniker<IWvListener> lreg("tcp", listener);
+
+
 WvTCPConn::WvTCPConn(const WvIPPortAddr &_remaddr)
 {
     remaddr = (_remaddr.is_zero() && FORCE_NONZERO)
@@ -119,7 +135,7 @@ WvTCPConn::WvTCPConn(WvStringParm _hostname, uint16_t _port)
 	do_connect();
     }
     else
-	dns.findaddr(0, hostname, NULL);
+	check_resolver();
 }
 
 
@@ -138,6 +154,7 @@ void WvTCPConn::nice_tcpopts()
     
     int value = 1;
     setsockopt(getfd(), SOL_SOCKET, SO_KEEPALIVE, &value, sizeof(value));
+    low_delay();
 }
 
 
@@ -176,8 +193,18 @@ void WvTCPConn::do_connect()
 	nice_tcpopts();
     }
     
-    sockaddr *sa = remaddr.sockaddr();
-    int ret = connect(getfd(), sa, remaddr.sockaddr_len()), err = errno;
+#ifndef _WIN32
+    WvIPPortAddr newaddr(remaddr);
+#else
+    // Win32 doesn't like to connect to 0.0.0.0:port; it means "any address
+    // on the local machine", so let's just force localhost
+    WvIPAddr zero;
+    WvIPPortAddr newaddr(WvIPAddr(remaddr)==zero
+			    ? WvIPAddr("127.0.0.1") : remaddr,
+			 remaddr.port);
+#endif
+    sockaddr *sa = newaddr.sockaddr();
+    int ret = connect(getfd(), sa, newaddr.sockaddr_len()), err = errno;
     assert(ret <= 0);
     
     if (ret == 0 || (ret < 0 && err == EISCONN))
@@ -209,6 +236,7 @@ void WvTCPConn::check_resolver()
     }
     else if (dnsres > 0)
     {
+	// fprintf(stderr, "%p: resolver succeeded!\n", this);
 	remaddr = WvIPPortAddr(*ipr, remaddr.port);
 	resolved = true;
 	do_connect();
@@ -255,7 +283,7 @@ void WvTCPConn::pre_select(SelectInfo &si)
     if (!resolved)
         dns.pre_select(hostname, si);
 
-    if (resolved && isok()) 
+    if (resolved) 
     {
 	bool oldw = si.wants.writable;
 	if (!isconnected()) {
@@ -352,25 +380,23 @@ size_t WvTCPConn::uwrite(const void *buf, size_t count)
 
 
 WvTCPListener::WvTCPListener(const WvIPPortAddr &_listenport)
-	: listenport(_listenport)
+	: WvListener(new WvFdStream(socket(PF_INET, SOCK_STREAM, 0)))
 {
+    WvFdStream *fds = (WvFdStream *)cloned;
     listenport = _listenport;
-    auto_list = NULL;
-    auto_userdata = NULL;
-    
     sockaddr *sa = listenport.sockaddr();
     
     int x = 1;
 
-    setfd(socket(PF_INET, SOCK_STREAM, 0));
-    set_close_on_exec(true);
-    set_nonblock(true);
+    fds->set_close_on_exec(true);
+    fds->set_nonblock(true);
     if (getfd() < 0
 	|| setsockopt(getfd(), SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x))
 	|| bind(getfd(), sa, listenport.sockaddr_len())
 	|| listen(getfd(), 5))
     {
 	seterr(errno);
+	return;
     }
     
     if (listenport.port == 0) // auto-select a port number
@@ -393,75 +419,46 @@ WvTCPListener::~WvTCPListener()
 }
 
 
-//#include <wvlog.h>
-void WvTCPListener::close()
-{
-    WvFDStream::close();
-/*    WvLog log("ZAP!");
-    
-    log("Closing TCP LISTENER at %s!!\n", listenport);
-    abort();*/
-}
-
-
-WvTCPConn *WvTCPListener::accept()
+IWvStream *WvTCPListener::accept()
 {
     struct sockaddr_in sin;
     socklen_t len = sizeof(sin);
-    int newfd;
-    WvTCPConn *ret;
+    
+    if (!isok()) return NULL;
 
-    newfd = ::accept(getfd(), (struct sockaddr *)&sin, &len);
-    ret = new WvTCPConn(newfd, WvIPPortAddr(&sin));
-    return ret;
+    int newfd = ::accept(getfd(), (struct sockaddr *)&sin, &len);
+    if (newfd >= 0)
+	return wrap(new WvTCPConn(newfd, WvIPPortAddr(&sin)));
+    else if (errno == EAGAIN || errno == EINTR)
+	return NULL; // this listener is doing weird stuff
+    else
+    {
+	seterr(errno);
+	return NULL;
+    }
 }
 
 
 void WvTCPListener::auto_accept(WvIStreamList *list,
-				WvStreamCallback callfunc, void *userdata)
+				wv::function<void(IWvStream*)> cb)
 {
-    auto_list = list;
-    auto_callback = callfunc;
-    auto_userdata = userdata;
-    setcallback(accept_callback, this);
+    onaccept(wv::bind(&WvTCPListener::accept_callback, this, list,
+			 cb, _1));
 }
 
-void WvTCPListener::auto_accept(WvStreamCallback callfunc, void *userdata)
+void WvTCPListener::auto_accept(wv::function<void(IWvStream*)> cb)
 {
-    auto_callback = callfunc;
-    auto_userdata = userdata;
-    setcallback(accept_global_callback, this);
+    auto_accept(&WvIStreamList::globallist, cb);
 }
 
 
-void WvTCPListener::accept_global_callback(WvStream &s, void *userdata)
+void WvTCPListener::accept_callback(WvIStreamList *list,
+				    wv::function<void(IWvStream*)> cb,
+				    IWvStream *_conn)
 {
-    WvTCPListener &l = *(WvTCPListener *)userdata; 
-    WvTCPConn *connection = l.accept();
-    connection->setcallback(l.auto_callback, l.auto_userdata);
-    WvIStreamList::globallist.append(connection, true, "WvTCPConn");
-}
-
-
-void WvTCPListener::accept_callback(WvStream &s, void *userdata)
-{
-    WvTCPListener &l = *(WvTCPListener *)userdata;
-
-    WvTCPConn *connection = l.accept();
-    connection->setcallback(l.auto_callback, l.auto_userdata);
-    l.auto_list->append(connection, true, "WvTCPConn");
-}
-
-
-size_t WvTCPListener::uread(void *, size_t)
-{
-    return 0;
-}
-
-
-size_t WvTCPListener::uwrite(const void *, size_t)
-{
-    return 0;
+    WvStreamClone *conn = new WvStreamClone(_conn);
+    conn->setcallback(wv::bind(cb, conn));
+    list->append(conn, true, "WvTCPConn");
 }
 
 
